import mistune
from jinja2 import contextfilter
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from slugify import slugify


class KartMistuneRenderer(mistune.HTMLRenderer):
    def __init__(self, context, *args, **kwargs):
        self.context = context
        super().__init__(*args, **kwargs)

    def link(self, link, text=None, title=None):
        if text is None:
            text = link
        a = f'<a href="{self.context["url"](link)}"'
        if title:
            a += f' title="{mistune.scanner.escape_html(title)}"'
        a += f">{text}</a>"
        return a

    def block_code(self, text, lang):
        if lang:
            lexer = get_lexer_by_name(lang, stripall=True)
            formatter = HtmlFormatter(wrapcode=True)
            return highlight(text, lexer, formatter)
        else:
            return f"<pre><code>{mistune.escape(text.strip())}</code></pre>\n"

    def heading(self, text, level, raw=None):
        return f"<h{level} id={slugify(text)}>{text}</h{level}>\n"


@contextfilter
def markdown_to_html(context, string):
    return mistune.Markdown(
        renderer=KartMistuneRenderer(context=context, escape=False),
        plugins=[
            mistune.plugins.plugin_strikethrough,
            mistune.plugins.plugin_table,
            mistune.plugins.plugin_task_lists,
        ],
    )(string)


class TocRenderer(mistune.renderers.BaseRenderer):
    def text(self, text):
        return text

    def heading(self, children, level):
        return {"title": children, "id": slugify(children), "level": level}

    def _get_method(self, name):
        if name in ["heading", "text"]:
            return super(TocRenderer, self)._get_method(name)

        def null(*args, **kwargs):
            return ""

        return null

    def finalize(self, data):
        data = list(data)
        if len(data) == 1:
            return data[0]
        return [x for x in data if x]


def markdown_to_toc(string):
    return mistune.Markdown(
        renderer=TocRenderer(),
    )(string)



import argparse
import fnmatch
import shutil
import sys
import threading
import traceback
from copy import deepcopy
from http.server import HTTPServer
from pathlib import Path

from kart.utils import KartMap, KartObserver, KartRequesHandler, merge_dicts


class Kart:
    def __init__(
        self,
        miners=[],
        content_modifiers=[],
        mappers=[],
        map_modifiers=[],
        renderers=[],
        config={},
        build_location="_site",
    ):
        self.miners = miners
        self.content_modifiers = content_modifiers
        self.mappers = mappers
        self.map_modifiers = map_modifiers
        self.renderers = renderers
        self.config = config
        self.build_location = Path(build_location)
        self.lock = threading.Lock()

    def check_config(self):
        default = {
            "name": "Example",
            "base_url": "https://example.org",
            "pagination": {"per_page": 5, "skip": 0},
            "timezone": "UTC",
        }
        merge_dicts(self.config, default)

    def mine_data(self, start=True):
        self.site = {"config": self.config}
        for miner in self.miners:
            if start:
                miner.read_data()
            self.site.update(miner.collect())
        for modifier in self.content_modifiers:
            modifier.modify(self.site)

    def create_map(self):
        self.map = KartMap(self.config["base_url"])
        for mapper in self.mappers:
            self.map.update(mapper.map(self.site))
        for modifier in self.map_modifiers:
            modifier.modify(self.map, self.site)

    def write(self):
        for renderer in self.renderers:
            renderer.render(self.map, self.site, self.build_location)

    def build(self):
        self.check_config()
        self.mine_data()
        self.create_map()
        shutil.rmtree(self.build_location, ignore_errors=True)
        self.build_location.mkdir(parents=True, exist_ok=True)
        self.write()

    # _site and _map are set and retrieved with a threading lock to prevent data races
    # _site and _map are set only when the creation of the map has finished
    # therefore it is not possible to access only partial data,
    # preventing errors when serving the site during development

    def update_data(self):
        self.check_config()
        self.mine_data(False)
        self.create_map()
        _site = deepcopy(self.site)
        _map = deepcopy(self.map)
        _urls = {}
        _regexes = {}
        for slug, page in self.map.items():
            _urls[page["url"]] = slug
            if "*" in page["url"] or "?" in page["url"]:
                _regexes[page["url"]] = slug
        with self.lock:
            self._site = _site
            self._map = _map
            self._urls = _urls
            self._regexes = _regexes

    def serve_page(self, handler, url):
        with self.lock:
            site_map = self._map
            urls = self._urls
            site = self._site
            regexes = self._regexes
        if url in urls:
            page = site_map[urls[url]]
        else:
            try:
                pattern = next((x for x in regexes if fnmatch.fnmatch(url, x)))
                page = site_map[regexes[pattern]]
            except StopIteration:
                page = None
        if page:
            self.renderer_dict[page["renderer"]].serve(handler, page, site_map, site)

    def serve(self, port=9000):
        self.renderer_dict = {}
        observer = KartObserver()
        observer.action = self.update_data
        for miner in self.miners:
            miner.start_watching(observer)
        observer.start()
        for renderer in self.renderers:
            self.renderer_dict[renderer.name] = renderer
            renderer.start_serving()
        handler_class = KartRequesHandler
        handler_class.action = self.serve_page
        httpd = HTTPServer(("", port), handler_class)
        httpd.timeout = 0.1
        self.update_data()
        shutil.rmtree(self.build_location, ignore_errors=True)

        while True:
            try:
                httpd.handle_request()
            except KeyboardInterrupt:
                print("\rexiting")
                for miner in self.miners:
                    miner.stop_watching()
                for renderer in self.renderers:
                    renderer.stop_serving()
                observer.stop()
                observer.join()
                sys.exit()
            except Exception:
                print(traceback.format_exc())

    def run(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "command", help="command to execute", choices={"build", "serve"}
        )
        parser.add_argument(
            "-p", "--port", help="port to bind to", default=9000, type=int
        )
        args = parser.parse_args()
        if args.command == "build":
            self.build()
        if args.command == "serve":
            self.config["base_url"] = f"http://localhost:{args.port}"
            self.serve(args.port)



from collections import OrderedDict
from pathlib import Path

from watchdog.events import RegexMatchingEventHandler

from kart.utils import slug_from_path

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


class Miner:
    def read_data(self):
        raise NotImplementedError

    def collect(self):
        raise NotImplementedError

    def start_watching(self):
        raise NotImplementedError

    def stop_watching(self):
        raise NotImplementedError


class DefaultMiner(Miner):
    def collect(self):
        raise NotImplementedError

    def collect_single_file(self, file):
        raise NotImplementedError

    def read_data(self):
        self.data = OrderedDict()
        for file in filter(Path.is_file, self.dir.iterdir()):
            object = self.collect_single_file(file)
            if object:
                self.data.update(object)

    def start_watching(self, observer):
        class Handler(RegexMatchingEventHandler):
            def on_moved(_, event):
                self.data.pop(slug_from_path(self.dir, Path(event.src_path)))
                self.data.update(self.collect_single_file(Path(event.src_path)))

            def on_modified(_, event):
                self.data.update(self.collect_single_file(Path(event.src_path)))

            def on_deleted(_, event):
                self.data.pop(slug_from_path(self.dir, Path(event.src_path)))

        self.read_data()
        observer.schedule(Handler(ignore_directories=True), self.dir, recursive=False)

    def stop_watching(self):
        pass


class DefaultMarkdownMiner(DefaultMiner):
    def collect_single_file(self, file):
        with file.open("r") as f:
            data = f.read().split("---")
            metadata = YamlLoader(data[1]).get_data()
            content = "---".join(data[2:])
            object = metadata
            slug = slug_from_path(self.dir, file)
            object["slug"] = slug
            object["content"] = content
            object["content_type"] = "markdown"
            if "draft" in object.keys():
                if object["draft"]:
                    return False
            return {slug: object}


class DefaultCollectionMiner(DefaultMarkdownMiner):
    def __init__(self, collection_name, directory="collections"):
        self.collection_name = collection_name
        self.dir = Path() / directory / collection_name

    def collect(self):
        return {self.collection_name: self.data}


class DefaultTaxonomyMiner(DefaultMarkdownMiner):
    def __init__(self, taxonomy_name, directory="taxonomies"):
        self.taxonomy_name = taxonomy_name
        self.dir = Path() / directory / taxonomy_name

    def collect(self):
        return {self.taxonomy_name: self.data}


class DefaultPageMiner(DefaultMarkdownMiner):
    def __init__(self, directory="pages"):
        self.dir = Path(directory)

    def collect(self):
        return {"pages": self.data}


class DefaultDataMiner(DefaultMiner):
    def __init__(self, directory="data"):
        self.dir = Path(directory)

    def collect_single_file(self, file):
        with file.open("r") as f:
            slug = slug_from_path(self.dir, file)
            return {slug: YamlLoader(f.read()).get_data()}

    def collect(self):
        return {"data": self.data}


class DefaultDocumentationMiner(DefaultMarkdownMiner):
    def __init__(self, directory="docs"):
        self.dir = Path(directory)

    def __recursive_read_data(self, dir, level=0):
        if dir.joinpath("navigation.yml").exists():
            nav_file = dir.joinpath("navigation.yml").open()
            nav_data = YamlLoader(nav_file.read()).get_data()
            paths = []
            for x in nav_data:
                if "page" in x.keys():
                    paths.append(dir.joinpath(x["page"]))
                elif "section" in x.keys():
                    paths.append(dir.joinpath(x["section"]))
        else:
            paths = dir.iterdir()
            nav_data = []

        for i, item in enumerate(paths):
            if item.is_file():
                object = self.collect_single_file(item)
                slug, page = list(object.items())[0]
                toc_entry = {"title": page["title"], "slug": slug, "level": level}
                self.docs_global_toc.append(toc_entry)
                if object:
                    self.markdown_data.update(object)
            elif item.is_dir():
                toc_entry = {"title": nav_data[i]["name"], "slug": None, "level": level}
                self.docs_global_toc.append(toc_entry)
                self.__recursive_read_data(item, level + 1)

    def read_data(self):
        self.markdown_data = OrderedDict()
        self.docs_global_toc = []
        self.__recursive_read_data(self.dir)

    def collect(self):
        return {"docs": self.markdown_data, "docs_global_toc": self.docs_global_toc}

    def start_watching(self, observer):
        class Handler(RegexMatchingEventHandler):
            def on_any_event(_, event):
                self.read_data()

        self.read_data()
        observer.schedule(Handler(), self.dir, recursive=True)



from kart.utils import paginate


class Mapper:
    def map(self, site):
        raise NotImplementedError


class RuleMapper(Mapper):
    def __init__(self, rules=[]):
        self.rules = rules  # a rule is a function

    def map(self, site):
        urls = {}
        for rule in self.rules:
            urls.update(rule(site))
        return urls


class ManualMapper(Mapper):
    def __init__(self, pages={}):
        self.pages = pages

    def map(self, site):
        return self.pages


class DefaultCollectionMapper(Mapper):
    def __init__(self, collection_name, template="item.html", base_url=""):
        self.template = template
        self.base_url = base_url
        self.collection_name = collection_name

    def map(self, site):
        urls = {}
        collection = site[self.collection_name]
        for object in collection.values():
            slug = f"{self.collection_name}.{object['slug']}"
            url = self.base_url + f"/{self.collection_name}/{object['slug']}/"
            if "template" in object:
                template = object["template"]
            else:
                template = self.template
            page = {
                "url": url,
                "data": object,
                "template": template,
                "renderer": "default_site_renderer",
            }
            urls[slug] = page
        return urls


class DefaultPageMapper(Mapper):
    def __init__(self, template="page.html"):
        self.template = template

    def map(self, site):
        urls = {}
        for slug in site["pages"]:
            page = site["pages"][slug]
            if "url" in page:
                url = page["url"]
            elif slug == "index":
                url = "/"
            else:
                url = f"/{slug}/"
            if "template" in page:
                template = page["template"]
            else:
                template = self.template
            page = {
                "url": url,
                "data": page,
                "template": template,
                "renderer": "default_site_renderer",
            }
            urls[slug] = page
        return urls


class DefaultIndexMapper(Mapper):
    def __init__(self, collection_name, template="index.html", base_url=""):
        self.template = template
        self.base_url = base_url
        self.collection_name = collection_name

    def map(self, site):
        items = list(site[self.collection_name].values())
        filtered_items = items[site["config"]["pagination"]["skip"] :]
        paginated_map = paginate(
            objects=filtered_items,
            per_page=site["config"]["pagination"]["per_page"],
            template=self.template,
            base_url=self.base_url + "/index/page",
            slug=f"{self.collection_name}_index",
            additional_data={},
        )
        paginated_map[f"{self.collection_name}_index.1"]["url"] = self.base_url + "/"
        return paginated_map


class DefaultTaxonomyMapper(Mapper):
    def __init__(
        self, collection_name, taxonomy_name, template="tag.html", base_url=""
    ):
        self.template = template
        self.base_url = base_url
        self.collection_name = collection_name
        self.taxonomy_name = taxonomy_name

    def map(self, site):
        urls = {}
        for taxonomy in site[self.taxonomy_name].values():
            slug = taxonomy["slug"]
            items = site[self.collection_name].values()
            filtered_items = []
            for item in items:
                if isinstance(item[self.taxonomy_name], str):
                    if item[self.taxonomy_name] == slug:
                        filtered_items.append(item)
                elif isinstance(item[self.taxonomy_name], list):
                    if slug in item[self.taxonomy_name]:
                        filtered_items.append(item)
            urls.update(
                paginate(
                    objects=filtered_items,
                    per_page=site["config"]["pagination"]["per_page"],
                    template=self.template,
                    base_url=self.base_url + f"/{self.taxonomy_name}/{slug}",
                    slug=f"{self.taxonomy_name}.{slug}",
                    additional_data=taxonomy,
                )
            )
        return urls


class DefaultFeedMapper(Mapper):
    def __init__(self, collections=[]):
        self.collections = collections

    def map(self, site):
        return {
            "feed": {
                "url": "/atom.xml",
                "data": {"collections": self.collections},
                "template": "",
                "renderer": "default_feed_renderer",
            }
        }


class DefaultDocumentationMapper(Mapper):
    def __init__(self, template="page.html", base_url=""):
        self.template = template
        self.base_url = base_url

    def map(self, site):
        urls = {}
        previous_slug = None
        for slug, page in site["docs"].items():
            if "url" in page:
                url = page["url"]
            elif slug == "index":
                url = "/"
            else:
                url = f"/{slug}/"
            if len(urls):
                page["previous_page"] = previous_slug
                urls[previous_slug]["data"]["next_page"] = slug
            previous_slug = slug
            map_page = {
                "url": self.base_url + url,
                "data": {**page},
                "template": self.template,
                "renderer": "default_site_renderer",
            }
            urls[slug] = map_page
        return urls



from collections import OrderedDict


class ContentModifier:
    def modify(self, site):
        raise NotImplementedError


class MapModifier:
    def modify(self, map, site):
        raise NotImplementedError


class RuleContentModifier(ContentModifier):
    def __init__(self, rules=[]):
        self.rules = rules  # a rule is a function

    def modify(self, site):
        for rule in self.rules:
            rule(site)


class RuleMapModifier(MapModifier):
    def __init__(self, rules=[]):
        self.rules = rules  # a rule is a function

    def modify(self, map, site):
        for rule in self.rules:
            rule(map, site)


class CollectionSorter:
    def __init__(self, collection_name, key, reverse=False):
        self.collection_name = collection_name
        self.key = key
        self.reverse = reverse

    def modify(self, site):
        data = site[self.collection_name]
        sorted_data = sorted(data.items(), key=lambda x: x[1][self.key])
        if self.reverse:
            sorted_data.reverse()
        site[self.collection_name] = OrderedDict(sorted_data)



import shutil
from datetime import datetime, time
from distutils.dir_util import copy_tree
from http.server import SimpleHTTPRequestHandler
from multiprocessing import Pool
from pathlib import Path

from dateutil import tz
from jinja2 import Environment, FileSystemLoader

from kart.markdown import markdown_to_html, markdown_to_toc
from kart.utils import date_to_string


class Renderer:
    def render(self, map, site, build_location):
        raise NotImplementedError

    def start_serving(self):
        pass

    def serve(self, http_handler, page, map, site):
        raise NotImplementedError

    def stop_serving(self):
        pass


class DefaultRenderer(Renderer):
    def render_single(self, map, site):
        raise NotImplementedError

    def render(self, map, site, build_location):
        for page in map.values():
            if page["renderer"] != self.name:
                continue
            rendered_file = self.render_single(page, map, site)
            path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as f:
                f.write(rendered_file)

    def serve(self, http_handler, page, map, site):
        http_handler.send_response(200)
        http_handler.send_header("Content-type", self.content_type)
        http_handler.end_headers()
        http_handler.wfile.write(bytes(self.render_single(page, map, site), "utf-8"))


class DefaultSiteRenderer(DefaultRenderer):
    def __init__(
        self,
        name="default_site_renderer",
        template_folder="templates",
        filters={
            "html": markdown_to_html,
            "toc": markdown_to_toc,
            "date_to_string": date_to_string,
        },
        process_count=1,
    ):
        self.name = name
        self.content_type = "text/html"
        self.template_folder = template_folder
        self.process_count = process_count
        self.env = Environment(loader=FileSystemLoader(self.template_folder))
        self.env.filters.update(filters)

    def render_single(self, page, map, site):
        if self.name != page["renderer"]:
            return

        template = self.env.get_template(page["template"])
        page = {**page["data"], "url": page["url"]}

        return template.render(page=page, site=site, url=map.url)

    def _render(self, key, map, site, build_location):
        page = map[key]
        if page["renderer"] != self.name:
            return
        rendered_file = self.render_single(page, map, site)
        path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
        path = path / "index.html"
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(rendered_file)

    def render(self, map, site, build_location="_site"):
        if self.process_count == 1:
            for key in map:
                self._render(key, map, site, build_location)
        else:
            with Pool(self.process_count) as p:
                p.starmap(
                    self._render, ((key, map, site, build_location) for key in map)
                )


class DefaultFeedRenderer(DefaultRenderer):
    def __init__(self, name="default_feed_renderer"):
        self.name = name
        self.content_type = "application/xml"

    def render_single(self, page, map, site):
        feed_entries = []
        for collection in page["data"]["collections"]:
            for object in site[collection].values():
                feed_entries.append([map.url(collection, object["slug"]), object])
        feed_entries.sort(key=lambda x: x[1]["date"], reverse=True)
        atom = '<feed xmlns="http://www.w3.org/2005/Atom">'
        atom += f'<id>"{map.url("/")}"</id>'
        atom += f"<title>{site['config']['name']}</title>"
        timezone = site["config"]["timezone"]
        updated_time = datetime.now().replace(tzinfo=tz.gettz(timezone))
        atom += f"<updated>{updated_time.isoformat()}</updated>"
        atom += f'<link href="{map.url("/")}"/>'
        atom += f'<link href="{map.url(page["url"])}" rel="self"/>'
        for url, entry in feed_entries:
            atom += "<entry>"
            atom += f"<id>{url}</id>"
            title = entry["title"] if "title" in entry.keys() else entry["name"]
            atom += f"<title>{title}</title>"
            entry_time = datetime.combine(entry["date"], time(12))
            entry_time.replace(tzinfo=tz.gettz(timezone))
            atom += f"<updated>{entry_time.isoformat()}</updated>"
            if "description" in entry.keys():
                atom += f'<summary>{entry["description"]}</summary>'
            atom += f'<link href="{url}" link="alternate"/>'
            atom += "</entry>"
        atom += "</feed>"
        return atom


class DefaultSitemapRenderer(DefaultRenderer):
    def __init__(self, name="default_sitemap_renderer"):
        self.name = name
        self.content_type = "application/xml"

    def render_single(self, page, map, site):
        base_url = site["config"]["base_url"]
        sitemap = '<?xml version="1.0" encoding="UTF-8"?>'
        sitemap += '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
        for x in map.values():
            if x["renderer"] != "default_site_renderer":
                continue
            sitemap += f"<url><loc>{base_url+x['url']}</loc></url>"
        sitemap += "</urlset>"
        return sitemap


class DefaultStaticFilesRenderer(Renderer):
    def __init__(self, name="default_static_files_renderer", directory="static"):
        self.name = name
        self.dir = directory

    def render(self, map, site, build_location):
        for page in map.values():
            if page["renderer"] != self.name:
                continue
            destination = Path(build_location) / Path(*Path(page["url"][:-1]).parts[1:])
            shutil.copytree(self.dir, destination)

    def serve(self, http_handler, page, map, site):
        return SimpleHTTPRequestHandler.do_GET(http_handler)


class DefaultRootDirRenderer(Renderer):
    def __init__(self, name="default_root_dir_renderer", directory="root"):
        self.name = name
        self.dir = directory

    def render(self, map, site, build_location):
        for page in map.values():
            if page["renderer"] != self.name:
                continue
            destination = Path(build_location) / Path(*Path(page["url"][:-1]).parts[1:])
            copy_tree(self.dir, str(destination))

    def serve(self, http_handler, page, map, site):
        http_handler.path = "/root" + http_handler.path
        return SimpleHTTPRequestHandler.do_GET(http_handler)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kart']

package_data = \
{'': ['*']}

install_requires = \
['jinja2>=2.11,<3.0',
 'mistune==2.0.0rc1',
 'pygments>=2.6.1,<3.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-slugify>=4.0.1,<5.0.0',
 'pyyaml>=5.1,<6.0',
 'watchdog>=0.10.3,<0.11.0']

setup_kwargs = {
    'name': 'kart',
    'version': '0.10.0',
    'description': 'A very flexible static site generator written in python',
    'long_description': '# Kart\nA very flexible static site generator written in python\n\n# Getting started\nInstall Kart with pip\n```bash\n$ pip install Kart\n```\n\nBuild the basic structure\n```bash\n$ python -m kart init\n```\n\nIn this configuration Kart will only build a basic blog with a paginated blog index and paginated tags. If you want to customize the urls of the blog you will have to modify main.py with custom python code\n\n\nYou can then build and serve your site with this command\n```bash\n$ python3 main.py serve\n```\n\n# Disclaimer\nKart is not yet ready to use in a real-world scenario. It is still in development its api can change abruptly each minor version.\n\nI am currently writing the [documentation](https://giacomocaironi.gitlab.io/Kart) of kart but it is by no mean complete. If you want to look at some examples you can look the docs folder, where the documentation is held, and the source code of the [example site](https://gitlab.com/giacomocaironi/Kart/-/tree/master/kart_quickstart)\n',
    'author': 'Giacomo Caironi',
    'author_email': 'giacomo.caironi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import math
import queue
import traceback
from collections import UserDict
from http.server import SimpleHTTPRequestHandler

from slugify import slugify
from watchdog.observers import Observer


class KartObserver(Observer):
    def run(self):
        while self.should_keep_running():
            try:
                self.dispatch_events(self.event_queue, self.timeout)
                if self.event_queue.empty():
                    self.action()
            except queue.Empty:
                pass
            except Exception:
                print(traceback.format_exc())


class KartRequesHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        try:
            self.action(self, self.path)
        except Exception:
            print(traceback.format_exc())


class KartMap(UserDict):
    def __init__(self, base_url, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_url = base_url

    def url(self, *name):
        name = ".".join(name)
        if not name:
            return ""
        if name in self.data.keys():
            return self.base_url + self.data[name]["url"]
        elif name + ".1" in self.data.keys():
            return self.base_url + self.data[name + ".1"]["url"]
        elif "http" in name:
            return name
        elif "/" in name:
            return self.base_url + name
        print("Invalid url")
        return ""


def paginate(objects, per_page, template, base_url, slug, additional_data={}):
    urls = {}
    paginated_objects = [
        objects[x * per_page : (x + 1) * per_page]
        for x in range(max(math.ceil(len(objects) / per_page), 1))
    ]
    for i, objects in enumerate(paginated_objects, 1):

        if i > 1:
            previous_page = f"{slug}.{i-1}"
        else:
            previous_page = ""

        if i < len(paginated_objects):
            next_page = f"{slug}.{i+1}"
        else:
            next_page = ""

        paginator = {
            "objects": objects,
            "index": i,
            "next_page": next_page,
            "previous_page": previous_page,
        }
        data = {"paginator": paginator}
        data.update(additional_data)
        urls[f"{slug}.{i}"] = {
            "url": base_url + f"/{i}/" if i > 1 else base_url,
            "data": data,
            "template": template,
            "renderer": "default_site_renderer",
        }
    return urls


def date_to_string(date):
    return date.strftime("%b %d, %Y")


def slug_from_path(base_dir, path):
    path = path.relative_to(base_dir)
    path = str(path)[: -len(path.suffix)]
    result = slugify(path)
    return path  # result


def merge_dicts(a, b):
    for key in b:
        if key not in a:
            a[key] = b[key]
        else:
            if isinstance(b[key], dict):
                if isinstance(a[key], dict):
                    merge_dicts(a[key], b[key])
                else:
                    raise Exception
